;	start.asm for the GBA thanx to Eloist (eloist@gmx.de)

	AREA my_1st_start, CODE, READONLY

	IMPORT |Image$$RO$$Limit| 	; End of ROM code (=start of ROM data)
	IMPORT |Image$$RW$$Base| 	; Base of RAM to initialize
	IMPORT |Image$$ZI$$Base| 	; Base and limit of area
	IMPORT |Image$$ZI$$Limit| 	; to zero initialize

	IMPORT C_Entry

Cpu_Mode_USR	EQU 	0x10
Cpu_Mode_IRQ 	EQU 	0x12
Cpu_Mode_SVC 	EQU 	0x13

Cpu_Stack_USR	EQU	0x03007F00	; GBA USR stack adress
Cpu_Stack_IRQ	EQU	0x03007FA0	; GBA IRQ stack adress
Cpu_Stack_SVC	EQU	0x03007FE0	; GBA SVC stack adress

REG_BASE		EQU	0x04000000  ; Registers
OFFSET_REG_IE		EQU	0x200	    ; Interrupt Enable
V_BLANK_INTR_FLAG	EQU	0x0001      ; V Blank Interrupt
H_BLANK_INTR_FLAG	EQU	0x0002      ; H Blank Interrupt
V_COUNT_INTR_FLAG	EQU	0x0004      ; V Counter Matching Interrupt
TIMER0_INTR_FLAG	EQU	0x0008      ; Timer 0 Interrupt
TIMER1_INTR_FLAG	EQU	0x0010      ; Timer 1 Interrupt
TIMER2_INTR_FLAG	EQU	0x0020      ; Timer 2 Interrupt
TIMER3_INTR_FLAG	EQU	0x0040      ; Timer 3 Interrupt
SIO_INTR_FLAG		EQU	0x0080      ; Serial Communication Interrupt
DMA0_INTR_FLAG		EQU	0x0100      ; DMA 0 Interrupt
DMA1_INTR_FLAG		EQU	0x0200      ; DMA 1 Interrupt
DMA2_INTR_FLAG		EQU	0x0400      ; DMA 2 Interrupt
DMA3_INTR_FLAG		EQU	0x0800      ; DMA 3 Interrupt
KEY_INTR_FLAG		EQU	0x1000      ; Key Interrupt
CASSETTE_INTR_FLAG	EQU	0x2000      ; Game Pak Interrupt

	ENTRY
	b 	start

	DCD 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
 	DCD 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
 	DCD 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

start
	mov 	r0, #Cpu_Mode_IRQ:OR:0x80:OR:0x40 	; Disable interrupts
	msr 	CPSR_c, r0				; Enable IRQ mode
	ldr 	r13, =Cpu_Stack_IRQ			; Setup IRQ stack pointer

	mov 	r0, #Cpu_Mode_SVC:OR:0x80:OR:0x40 	; Disable interrupts
	msr 	CPSR_c, r0				; Enable SVC mode
	ldr 	r13, =Cpu_Stack_SVC			; Setup SVC stack pointer

	ldr 	r0, =|Image$$RO$$Limit| 	; Get pointer to ROM data
	ldr 	r1, =|Image$$RW$$Base| 		; and RAM copy
	ldr 	r3, =|Image$$ZI$$Base| 		; Zero init base => top of initialized data
	cmp 	r0, r1 				; Check that they are different
	beq 	%F1
0 	cmp 	r1, r3 				; Copy init data
	ldrcc 	r2, [r0], #4
	strcc 	r2, [r1], #4
	bcc	%B0
1	ldr 	r1, =|Image$$ZI$$Limit| 	; Top of zero init segment
	mov 	r2, #0
2 	cmp 	r3, r1 				; Zero init
	strcc 	r2, [r3], #4
	bcc	%B2

	mov 	r0, #Cpu_Mode_SVC:OR:0x40	; Only IRQ enabled
	msr 	CPSR_c, r0

	mov 	r0, #Cpu_Mode_USR:OR:0x80:OR:0x40	;
	msr 	CPSR_c, r0				; Switch to USR mode
	ldr 	sp, =Cpu_Stack_USR			; Setup USR stack pointer

	b	C_Entry		; Now branch to start of C code

final_return_loop
	b 	final_return_loop		; In case of a return...

;--------------------------------------------------------------------
;-         Interrupt Branch Process (Draw from Table) 32Bit  23-61c -
;--------------------------------------------------------------------
    EXTERN      IntrTable
    GLOBAL      intr_main
    ALIGN
    CODE32
intr_main
        mov     r3, #REG_BASE           ; Check IE/IF
        add     r3, r3, #OFFSET_REG_IE  ; r3: REG_IE
        ldr     r2, [r3]
        and     r1, r2, r2, lsr #16 	; r1: IE & IF
        mov     r2, #0
        ands    r0, r1, #V_BLANK_INTR_FLAG   ; V Blank interrupt
        bne     jump_intr
        add     r2, r2, #4
        ands    r0, r1, #H_BLANK_INTR_FLAG   ; H Blank interrupt
        bne     jump_intr
        add     r2, r2, #4
        ands    r0, r1, #V_COUNT_INTR_FLAG   ; V Blank interrupt
        bne     jump_intr
        add     r2, r2, #4
        ands    r0, r1, #TIMER0_INTR_FLAG    ; Timer 0 interrupt
        bne     jump_intr
        add     r2, r2, #4
        ands    r0, r1, #TIMER1_INTR_FLAG    ; Timer 1 interrupt
        bne     jump_intr
        add     r2, r2, #4
        ands    r0, r1, #TIMER2_INTR_FLAG    ; Timer 2 interrupt
        bne     jump_intr
        add     r2, r2, #4
        ands    r0, r1, #TIMER3_INTR_FLAG    ; Timer 3 interrupt
        bne     jump_intr
        add     r2, r2, #4
        ands    r0, r1, #SIO_INTR_FLAG       ; Serial Communication interrupt
        bne     jump_intr
        add     r2, r2, #4
        ands    r0, r1, #DMA0_INTR_FLAG      ; DMA0 interrupt
        bne     jump_intr
        add     r2, r2, #4
        ands    r0, r1, #DMA1_INTR_FLAG      ; DMA1 interrupt
        bne     jump_intr
        add     r2, r2, #4
        ands    r0, r1, #DMA2_INTR_FLAG      ; DMA2 interrupt
        bne     jump_intr
        add     r2, r2, #4
        ands    r0, r1, #DMA3_INTR_FLAG      ; DMA3 interrupt
        bne     jump_intr
        add     r2, r2, #4
        ands    r0, r1, #KEY_INTR_FLAG       ; Key interrupt
        bne     jump_intr
        add     r2, r2, #4
        ands    r0, r1, #CASSETTE_INTR_FLAG  ; Cassette interrupt
jump_intr
        strh    r0, [r3, #2]            ; IF Clear           11c
        ldr     r1, =IntrTable          ; Jump to User IRQ Process
        add     r1, r1, r2
        ldr     r0, [r1]
        bx      r0

	END

