//----------------------------------//
// Ghost AI by Gibby. 				//
// Thanks to Dr Scutt for the help. //
//----------------------------------//

#include "types.h"
#include "main.h"
#include "ghost.h"
#include "sprite.h"
#include "libc.h"
#include "data.h"

#define BOX_LEFT	100
#define BOX_RIGHT	124
#define BOX_OUT 	84
#define BLUE_SPEED	0x8000
#define SEEK_DIST	24

int 	GetRoute(short boxnum, GHOSTS *g, int ai, int gx, int gy);
void 	AddBoxToUnexplored(short boxnum,GHOSTS *g, char firsttime);
void 	DoGhostInBox(GHOSTS *g);
int 	GetGhostInBox(GHOSTS *g);
short 	GetCutoffBoxNumber(short boxnum);

short 	Unexplored[512];
short 	BoxUDone[512];
short 	CurrentExplored;
short 	UnexploredCount;
u8 	*joverlaps = (u8 *) &_joverlaps;
u8 	*junctions = (u8 *) &_junctions;

short 	PMBox;

short	*boxes = (short*)0x3003000;
short	*overlaps = (short*)0x3003800;

char 	RDirs[4] = {	PM_DOWN,PM_UP,PM_RIGHT,PM_LEFT	};	// Table of reverse directions (to avoid horrible IF statement).

int		gvel[4];

GHOSTS 	ghost[4];	// Struct array for ghosties.
short 	ExtraBoxes[5];	// Box numbers of 4 pills + fruit bonus.


//********************
//                   *
//	 START OF CODE   *
//                   *
//********************

//
// Seperate ghost movement routines because the cage in the middle isn't the same format as the rest of the maze.
//

void DoGhostInBox(GHOSTS *g)
{
	if (g->dir == PM_LEFT)
	{
		g->x-=HALF;
		if (POS(g->x) == BOX_LEFT)
			g->dir = PM_RIGHT;
	}
	else if (g->dir == PM_RIGHT)
	{
		g->x+=HALF;
		if (POS(g->x) == BOX_RIGHT)
			g->dir = PM_LEFT;
	}
	else if (g->dir == PM_UP)
	{
		g->y-=HALF;
		if (POS(g->y) == BOX_OUT)
		{
			g->InBox=1;	// Out of box, so let AI know.
			return;
		}
	}

	if (g->BlueTimer == 0)
	{
		if (g->Timer)
			g->Timer--;
		else
		{
			if (POS(g->x) == 112)
				g->dir = PM_UP;
		}
	}
	else
		g->BlueTimer--;
}

//
// Routine to intercept the ghost eyes and force them into the middle box.
//

int GetGhostInBox(GHOSTS *g)
{
	if (!(g->Flags&ENTERBOX))
	{
		if (POS(g->y) == BOX_OUT && POS(g->x) > 100 && POS(g->x) < 124)
		{
			g->Flags|=ENTERBOX;
			return 1;
		}
		return 0;
	}

	if (POS(g->x) < 112)
	{
		g->dir = PM_RIGHT;
		g->x+=UNIT;
		return 1;
	}
	else if (POS(g->x) > 112)
	{
		g->dir = PM_LEFT;
		g->x-=UNIT;
		return 1;
	}
	else if (POS(g->y)<108)
	{
		g->dir = PM_DOWN;
		g->y+=UNIT;
		return 1;
	}

	g->InBox = 2;
	g->Timer = 60*3;	// 3 second delay before coming out of box.
	g->dir = PM_LEFT;
	g->Flags &= ~(EATEN|ENTERBOX);

	return 1;
}

void UpdateGhost(GHOSTS *g, int ghostnum)
{
	int	u=0,d=0,l=0,r=0,AtJunction=0;
	int	Ddir;
	int	lrx,lry;
	int	udx,udy;

	if (g->InBox==2)
	{
		DoGhostInBox(g);
		return;
	}

	if (g->Flags&EATEN)
	{
		if (GetGhostInBox(g))
			return;
	}

//
// Dirty detection code because of the weird way the map is laid out. (Ripped from MAME version of Pacman).
//

	if (g->InBox==0)
	{
		lrx = POS(g->x)+4;
		lry = POS(g->y)+8;
		udx = POS(g->x)+8;
		udy = POS(g->y)+4;

		if ((POS(g->y)&7)==4)	// To make sure it's on a char boundary.
		{
			if (Map[((lrx+7+1)>>3)+((lry>>3)<<5)] < 7)	// Can he go right?
				r=1;
			if (Map[((lrx-1)>>3)+((lry>>3)<<5)] < 7)	// Can he go left?
				l=1;
		}

		if ((POS(g->x)&7)==4) 	// To make sure it's on a char boundary.
		{
			if (Map[(udx>>3)+(((udy+7+1)>>3)<<5)] < 7)	// Can he go down?
				d=1;
			if (Map[(udx>>3)+(((udy-1)>>3)<<5)] < 7)	// Can he go up?
				u=1;
		}

		if ((u|d) == 1 && (l|r) == 1)	// At a junction.
			AtJunction=1;
	}
	else
	{
		l=r=1;
		u=d=0;
		if (GetRandom()&1)
			g->dir=PM_LEFT;
		else
			g->dir=PM_RIGHT;
	}

	if (POS(g->x)<0)
		lrx=0;
	else if (POS(g->x)>=232)
		lrx=232>>3;
	else
		lrx = (POS(g->x)+8)>>3;
	lry = ((POS(g->y)+8)>>3)<<5;

	if (AtJunction || (g->Flags & QUICK_TURN))
	{
		Ddir = GetRoute(boxes[lrx+lry], g, 1, lrx, lry);	// Get destination direction.
		g->Flags &= ~QUICK_TURN;
	}
	else
	{
		GetRoute(boxes[lrx+lry], g, 0,0,0);	// Just do detection.
		Ddir = g->dir;
	}

	if ((Ddir==PM_LEFT && l) ||
		(Ddir==PM_RIGHT && r) ||
		(Ddir==PM_UP && u) ||
		(Ddir==PM_DOWN && d))
	{
		if (g->InBox==1)	// Just got out of box.
		{
			g->InBox=0;
			g->dir=8;
		}
		SetGhostVel(g,ghostnum,Ddir);
	}

	AccelerateGhosts();
	g->x+=g->xv;
	g->y+=g->yv;

	if (POS(g->x)<=-16)
		g->x = TOFIXED(239);
	else if (POS(g->x)>=240)
		g->x = TOFIXED(-15);

	if (g->BlueTimer)
	{
		g->BlueTimer--;
		if (!g->BlueTimer)
			g->Flags |= QUICK_TURN;
	}
}

// Don't forget to slow them down round corners.

int GetRoute(short boxnum, GHOSTS *g, int ai, int gx, int gy)
{
	short GBox=boxnum;
	short tbox;
	int	lp,dir;
	u8	junc;
	int	px;

	px=pacman.x;
	if (POS(px)<0)		// Cater for tunnel left.
	{
		if (pacman.dir==PM_LEFT)
			px = TOFIXED(224);
		else
			px = TOFIXED(0);
	}
	else if (POS(px)>224)		// Cater for tunnel right.
	{
		if (pacman.dir==PM_RIGHT)
			px = TOFIXED(0);
		else
			px = TOFIXED(224);
	}

	g->Boxnum = boxnum;

	if (g->Flags&EATEN)
	{
		if (ai==0)
			return 0;
		PMBox = boxes[(116>>3)+((108>>3)<<5)];	// Middle box box number.
	}
	else
	{
		int	dx,dy;

		PMBox = boxes[((POS(px)+8)>>3)+(((POS(pacman.y)+8)>>3)<<5)];	// Pacman's box number.
		dx = (POS(g->x)+5) - (POS(px)+5);
		dy = (POS(g->y)+5) - (POS(pacman.y)+5);
		dx = ABS(dx);
		dy = ABS(dy);

		if (PMBox == boxnum || (dx<6&&dy<6))
		{
			if (g->BlueTimer)
			{
				g->Flags|=EATEN;
				g->Flags|=QUICK_TURN;
				g->BlueTimer=0;
				DoGhostScore(g,GHOST200+GhostScore);
				Score+=200<<GhostScore;
				if (GhostScore<3)
					GhostScore++;
			}
			else
			{
				pacman.Timer=0;	// initiate death sequence.
			}
			ai=1;
		}

		if (ai==0)
			return 0;

		if (!g->BlueTimer)
		{
			if (g->AITime)
			{
				g->AITime--;
				if (!g->AITime)	// Toggle AI type between roamer and its normal AI.
				{
					g->AITime = 3*60+(GetRandom()&255);	// 3-7 seconds AI toggle time.

					if (g->CurrentAI != g->NormalAI)
					{
						if ((GetRandom()&7)==0)
							g->CurrentAI = ROAMER;
						else
							g->CurrentAI = g->NormalAI;

						g->AITime += (GetRandom()&127);	// Longer time in natural ghost AI mode.
					}
					else
					{
						g->CurrentAI = ROAMER;
					}
				}
				else
				{
					if ((GetRandom()&31)==0)
					{
						if (g->NormalAI == CUTOFF)
							g->CurrentAI = CUTOFF;
						else
							g->CurrentAI = PMHUNTER;
					}
				}
			}

			switch (g->CurrentAI)
			{
				case ROAMER:
					do
					{
						PMBox = GetRandom()%NUM_BOXES;
					}
					while (PMBox == boxnum);

					break;

				case PILLHUNTER:
					if (POS(px)<120)
					{
						if (POS(pacman.y)<112)
							PMBox = ExtraBoxes[0];	// Top left pill.
						else
							PMBox = ExtraBoxes[2];	// Bottom left pill.
					}
					else
					{
						if (POS(pacman.y)<112)
							PMBox = ExtraBoxes[1];	// Top right pill.
						else
							PMBox = ExtraBoxes[3];	// Bottom right pill.
					}
					break;

				case CUTOFF:
					PMBox = boxes[((POS(px)+8)>>3)+(((POS(pacman.y)+8)>>3)<<5)];	// Pacman's box number.
					PMBox = GetCutoffBoxNumber(PMBox);
					break;
				default:	//  PMHUNTER
					PMBox = boxes[((POS(px)+8)>>3)+(((POS(pacman.y)+8)>>3)<<5)];	// Pacman's box number.
					break;
			}
		}
	}

	junc = junctions[gx+gy];

	if (junc==0xff)	// Not at a junction?
	{
		if ((g->Flags&QUICK_TURN) && (GetRandom()&1))
			dir=RDirs[g->dir];
		else
			dir=g->dir;
	}
	else
		dir = joverlaps[(junc*NUM_BOXES)+PMBox];

	if ((g->Flags&QUICK_TURN)==0 && dir == RDirs[g->dir])
	{
	 	for (lp=0;lp<4;lp++)
		{
			if (overlaps[(GBox<<2)+lp] != -1 && lp != g->dir && lp!=dir)
		 		break;
		}

		if (lp==4)
			dir = g->dir;
		else
			dir=lp;
	}

	if (g->BlueTimer)	// Run away.
	{
		char dirs[4],ndirs=0;

		for (lp=0;lp<4;lp++)
		{
			if (overlaps[(GBox<<2)+lp] != tbox && overlaps[(GBox<<2)+lp] != -1)
			{
				if (lp != RDirs[g->dir] || (g->Flags&QUICK_TURN))
					dirs[ndirs++]=lp;	// Get list of possible directions.
			}
		}

		if (ndirs)
			dir = dirs[GetRandom()%ndirs];
	}

	return dir;
}

void AddBoxToUnexplored(short boxnum,GHOSTS *g, char firsttime)
{
	int	lp;
	short *optr,*bptr,box;

	optr = overlaps;
	optr += (boxnum<<2);

	for (lp=0;lp<4;lp++,optr++)
	{
		box = *optr;
		bptr = &BoxUDone[box];

		if (box != -1 && bptr[0] < 0)
		{
			Unexplored[UnexploredCount++] = box; 	// Add to unexplored list.
			bptr[0]=boxnum;
		}
	}
}

short GetCutoffBoxNumber(short boxnum)
{
	int	lp,olp;
	short	dir;
	short *optr,box;

	dir = pacman.dir;

	for (lp=0;lp<SEEK_DIST;lp++)
	{
		optr = overlaps;
		optr += (boxnum<<2);

		if (optr[dir] != -1)
			boxnum = optr[dir];
		else
		{
			for (olp=0;olp<4;olp++)
			{
				if (optr[olp]!=-1)
					box++;
			}

			if (box > 2)	// Choice of directions so stop here.
				break;

			for (olp=0;olp<4;olp++)
			{
				if (optr[olp] != -1 && olp != RDirs[dir])
				{
					boxnum = optr[olp];
					break;
				}
			}
		}
	}

	return (boxnum);
}


void SetGhostVel(GHOSTS *g, int ghostnum, int dir)
{
	int	newdir=0;

	if (g->dir!=dir)
		newdir=1;
	g->dir = dir;

	if (dir == PM_LEFT)
	{
		g->yv = g->dyv = 0;
		if (g->Flags&EATEN)
			g->xv = g->dxv = -UNIT;

		if (newdir)
		{
			if (g->BlueTimer)
				g->dxv = -(gvel[ghostnum]-BLUE_SPEED);
			else
				g->dxv = -gvel[ghostnum];
			g->xv = g->dxv>>2;
			g->x &= ~0xffff;
		}
	}
	else if (dir == PM_RIGHT)
	{
		g->yv = g->dyv = 0;
		if (g->Flags&EATEN)
			g->xv = g->dxv = UNIT;
		if (newdir)
		{
			if (g->BlueTimer)
				g->dxv = (gvel[ghostnum]-BLUE_SPEED);
			else
				g->dxv = gvel[ghostnum];

			g->xv = g->dxv>>2;
			g->x&=~0xffff;
			g->x+=UNIT-g->xv;
		}
	}
	else if (dir == PM_UP)
	{
		g->xv = g->dxv = 0;
		if (g->Flags&EATEN)
			g->yv = g->dyv = -UNIT;

		if (newdir)
		{
			if (g->BlueTimer)
				g->dyv = -(gvel[ghostnum]-BLUE_SPEED);
			else
				g->dyv = -gvel[ghostnum];

			g->yv = g->dyv>>2;
			g->y&=~0xffff;
		}
	}
	else
	{
		g->xv = g->dxv = 0;
		if (g->Flags&EATEN)
			g->yv = g->dyv = UNIT;

		if (newdir)
		{
			if (g->BlueTimer)
				g->dyv = (gvel[ghostnum]-BLUE_SPEED);
			else
				g->dyv = gvel[ghostnum];

			g->yv = g->dyv>>2;
			g->y&=~0xffff;
			g->y+=UNIT-g->yv;
		}
	}
}

void AccelerateGhosts()
{
	GHOSTS *g;
	int	i;

	g=&ghost[0];
	for (i=0;i<4;i++)
	{
		g->xv += (g->dxv-g->xv)>>4;
		g->yv += (g->dyv-g->yv)>>4;
		g++;
	}
}

