//-------------------------//
// Pacman Advance by Gibby //
//-------------------------//

#include "types.h"
#include "title.h"
#include "main.h"
#include "ghost.h"
#include "data.h"
#include "libc.h"
#include "cosin.h"
#include "sprite.h"

extern	void	init_irq_cack(void);

void	download_sprite_data(void);
void	SetSpritePacman(void);
void 	SetSpriteGhost(void);
void 	SetSpriteBonus(void);
void 	UpdatePacman(void);
void	AnimatePacMan(void);
void 	InitNewGame(void);
void 	InitNewLevel(void);
void 	InitLife(void);
void 	Startup(void);
void 	DisplayScore(void);
void	ClearRemainingTextSprites(void);

GHOSTS 	pacman;		// Struct for Pacman.
SCORES 	scorespr;	// Struct array for extra sprites (scores, bonuses, etc.)

volatile u16	vblcounter;

short 	FadeSrc,FadeDst,FadeSpeed,FadeScreens;

short 	RawEdge=0,RawPad=0;

u8 		*tiles=(u8*)0x6004000;	// Vram character data address
u16		*m0=(u16*)0x6000000;	// Vram bkg map address.
u16		*tmp_pic;

u8 		Map[1024];				// Map of walls and pills. (8==wall, 1==small pill, 3==power pill. )

short 	CurrentLevel,GhostLevel;
short 	BonusTimer;

int 	Score;
int 	HighScore=10000;
int		ELScore;
int		GhostScore=0;
int		TxtSpr;
int		ExitLevel=0;

int		ScrollPos;		// Vertical scroll position (Whole map doesn't fit on screen. Doh!)
int		jlr=0,jud=0;	// Pacman desired direction set up from joypad.

int		P1Timer,ReadyTimer,PillsEaten;

short 	BonusScores[8] = {	100,300,500,700,1000,2000,3000,5000	};

// Anim frames and flip values.

u16		PMFrames[4][5] = 	{	{	2,3,4,3,0x2000	},	// Up.
					 		{	2,3,4,3,0		},	// Down.
		   			 		{	0,1,4,1,0x1000	},	// Left.
		   			 		{	0,1,4,1,0		}	// Right.
						};

// X and Y offsets to cater for alignment problem when Pacman is flipped
//						(noticed on the arcade version that they didn't bother fixing this!).

short	PMOffsets[4][2] = 	{	{ 0,-1 },	// Up.
				  				{ 0,0 },  	// Down.
				   				{ -1,0 }, 	// Left.
				   				{ 0,0 }   	// Right.
				   			};

char 	PillAnims[4] = 	{ 1,2,3,2	};		// Char values for power pill anim.

//********************
//                   *
//	 START OF CODE   *
//                   *
//********************

void Startup(void)
{
	short 	*sptr,*dptr,x;

	init_irq_cack();

	BG0CNT = 0x0007;		//turn on BG0CNT.  Also tells where the map and tile locations are for this bkg
	DISPCNT = 0x1140;		//sets the display mode to display BKG0 & OBJ.

	sptr = (short *)&_Boxes;	// Copy data into fast ram. (0x3000000 32k)
	dptr = (short *)boxes;
	for (x=0;x<1024;x++)
		*dptr++=*sptr++;

	sptr = (short *)&_Overlaps;	// Copy data into fast ram.
	dptr = overlaps;
	for (x=0;x<1288;x++)
		*dptr++=*sptr++;


}

void InitNewGame(void)
{
	int	x,y,eb=0;

	SetPaletteChr(&_pal,0,16);	// Upload palette into palette ram.

	DMA3Call((u32)&_chr,(u32)tiles,0x84000200);

	download_sprite_data();

	vblcounter=0;

	pacman.Lives=2;

	Score=0;
	ELScore=10000;

	tmp_pic = (u16*)(&_pic);
	for (y=0;y<32;y++)
	{
		for (x=0;x<30;x++)
		{
			if (*(tmp_pic++)&1023 == 3)
				ExtraBoxes[eb++] = boxes[(y*32)+x];	// Box number of power pills.
		}
	}

	ExtraBoxes[eb] = boxes[((140>>3)*32)+(120>>3)];	// Box number of bonus fruits.
	CurrentLevel=-1;
	GhostLevel=0;
}

void InitNewLevel(void)
{
	u8	*smptr,*dmptr;
	u16 	*tmpme;
	u16 	x,y;

	gvel[0]=G0VEL+(GhostLevel<<8);
	gvel[1]=G1VEL+(GhostLevel<<8);
	gvel[2]=G2VEL+(GhostLevel<<8);
	gvel[3]=G3VEL+(GhostLevel<<8);

	PillsEaten=0;

	tmp_pic = (u16*)(&_pic);		//have tmp_pic point to our .pic defined in data.asm
	tmpme = m0;
	for (y=0;y<32;y++) 	  		//loop to load in the map data
	{
		for (x=0;x<30;x++)
			*(tmpme++) = *(tmp_pic++);
		*(tmpme++) = 0;
		*(tmpme++) = 0;	// Right edge is blank (could have done it on the picture but couldn't be arsed!)
	}

	smptr = (u8*)&_DetMap;
	dmptr = (u8*)&Map[0];
	for (y=0;y<32;y++)
	{
		for (x=0;x<32;x++)
		{
			*dmptr++=*smptr++;
		}
	}

	if (CurrentLevel < 7)
		CurrentLevel++;

	if (GhostLevel < 24)
		GhostLevel++;
}

void InitLife(void)
{
	int x;

	ghost[0].x = TOFIXED(112);
	ghost[0].y = TOFIXED(108);
	ghost[0].xv = 0;
	ghost[0].yv = 0;
	ghost[0].dir=PM_UP;
	ghost[0].InBox=2;
	ghost[0].Timer=0;	// This ghost comes straight out of box.
	ghost[0].NormalAI=PMHUNTER;
	ghost[0].CurrentAI=ROAMER;
	ghost[0].AIRand = 31;	// How often it goes from roam to hunt if too far away to hunt directly.
	ghost[0].AITime= 2*60;	// How long it hunts if too far away to hunt directly.
	ghost[0].BlueTimer=0;
	ghost[0].Flags=0;

	ghost[1].x = TOFIXED(100);
	ghost[1].y = TOFIXED(108);
	ghost[1].xv = 0;
	ghost[1].yv = 0;
	ghost[1].dir= PM_RIGHT;
	ghost[1].InBox=2;
	ghost[1].Timer=60*1;	// 1 second delay before coming out of box.
	ghost[1].NormalAI=CUTOFF;
	ghost[1].CurrentAI=ROAMER;
	ghost[1].AIRand = 31;
	ghost[1].AITime= 2*60;	// How long it hunts if too far away to hunt directly.
	ghost[1].BlueTimer=0;
	ghost[1].Flags=0;

	ghost[2].x = TOFIXED(116);
	ghost[2].y = TOFIXED(108);
	ghost[2].xv = 0;
	ghost[2].yv = 0;
	ghost[2].dir= PM_RIGHT;
	ghost[2].InBox=2;
	ghost[2].Timer=60*2;	// 2 second delay before coming out of box.
	ghost[2].NormalAI=PILLHUNTER;
	ghost[2].CurrentAI=ROAMER;
	ghost[2].AIRand = 31;
	ghost[2].AITime= 2*60;	// How long it hunts if too far away to hunt directly.
	ghost[2].BlueTimer=0;
	ghost[2].Flags=0;

	ghost[3].x = TOFIXED(124);
	ghost[3].y = TOFIXED(108);
	ghost[3].xv = 0;
	ghost[3].yv = 0;
	ghost[3].dir= PM_LEFT;
	ghost[3].InBox=2;
	ghost[3].Timer=60*3;	// 3 second delay before coming out of box.
	ghost[3].NormalAI=ROAMER;
	ghost[3].CurrentAI=ROAMER;
	ghost[3].AIRand = 31;
	ghost[3].AITime= 2*60;	// How long it hunts if too far away to hunt directly.
	ghost[3].BlueTimer=0;
	ghost[3].Flags=0;

	pacman.x = TOFIXED(112);
	pacman.y = TOFIXED(180);
	pacman.xv = 0;
	pacman.yv = 0;
	pacman.def = 2<<1;
	pacman.Timer=-1;	// Death flag/count.
	jlr=-1;	// Make it go left from start.

	for (x=0;x<128;x++)	// Remove all the sprites from the screen.
		OAM[x].Attrib0 = 160;	// Setup Y off screen.

	scorespr.timer=-1;

	P1Timer=2*60;
	ReadyTimer=1*60;
	BonusTimer=30*60+(GetRandom()&1023);	// Between 30 and 60 seconds bonus repop.
}

void C_Entry()
{
	Startup();	// Setup irq's, download palettes and graphics, etc.

	while (1)
	{
		DoTitle();

		InitNewGame();

		ExitLevel=NEXT_LEVEL;

		while (pacman.Lives >= 0)
		{
			if (ExitLevel==NEXT_LEVEL)
				InitNewLevel();

			InitLife();

			ExitLevel=0;

			FadeDst=128;

			while (!ExitLevel)									//loop
			{
				ScrollPos = POS(pacman.y)-72;
				if (ScrollPos < 0)
					ScrollPos=0;
				else if (ScrollPos > 96)
					ScrollPos = 96;

				VSync(1);

				//ProfileRGB(255,0,0);
				Fade();

				TxtSpr=127;	// Reset printf sprites.

				BG0VOFS = ScrollPos;
				SetupBonus();
				SetSpritePacman();		// Draw Pacman.
				SetSpriteGhost();	// Draw the ghosts.
				SetSpriteBonus();	// Draw bonuses.

				if (Score>HighScore)
					HighScore=Score;
				if (Score>ELScore)
				{
					pacman.Lives++;
					ELScore+=20000;
				}

				DisplayScore();

				ReadJoypad();								//set our variable btn to be the value in the joypad register

				if (RawPad & J_LEFT)
					jlr=-1;
				else if (RawPad & J_RIGHT)
					jlr=1;
				if (RawPad & J_UP)
					jud=-1;
				else if (RawPad & J_DOWN)
					jud=1;

				if (RawEdge & J_START)
					PauseGame();

				if (P1Timer==0 && ReadyTimer==0)
				{
					if (pacman.Timer == -1)
						UpdatePacman();

					if (pacman.Timer == -1)
					{
						UpdateGhost(&ghost[0],0);
 						UpdateGhost(&ghost[1],1);
 						UpdateGhost(&ghost[2],2);
 						UpdateGhost(&ghost[3],3);

						if (ghost[0].BlueTimer==0 &&
							ghost[1].BlueTimer==0 &&
							ghost[2].BlueTimer==0 &&
							ghost[3].BlueTimer==0)	// Are any ghosts blue?
						{
							GhostScore=0;	// Reset eaten-ghost score.
						}
					}
				}

				AnimatePacMan();

				ClearRemainingTextSprites();

				//ProfileRGB(0,0,0);
			}

			if (ExitLevel==DEATH)
			{
				pacman.Lives--;
				if (pacman.Lives>=0)
				{
					FadeDst=0;	// FadeDown.
					WaitFade();
				}
			}

			if (ExitLevel==NEXT_LEVEL)
			{
				FinishLevel();
				FadeDst=255;
				WaitFade();
			}
		}

		GameOver();
	}
}


void UpdatePacman()
{
	int	u=0,d=0,l=0,r=0;
	int	lrx,lry;
	int	udx,udy;

// Dirty detection code because of the obscure size of the pacman
// and the weird way the map is laid out. (Ripped from MAME version of Pacman).

	lrx = POS(pacman.x)+4;
	lry = POS(pacman.y)+8;
	udx = POS(pacman.x)+8;
	udy = POS(pacman.y)+4;

	if ((POS(pacman.y)&7)==4)	// To make sure it's on a char boundary.
	{
		if (Map[((lrx+7+1)>>3)+((lry>>3)<<5)] < 7)	// Can he go right?
			r=1;
		if (Map[((lrx-1)>>3)+((lry>>3)<<5)] < 7)	// Can he go left?
			l=1;
	}

	if ((POS(pacman.x)&7)==4)	// To make sure it's on a char boundary.
	{
		if (Map[(udx>>3)+(((udy+7+1)>>3)<<5)] < 7)	// Can he go down?
			d=1;
		if (Map[(udx>>3)+(((udy-1)>>3)<<5)] < 7)	// Can he go up?
			u=1;
	}


	if (l==0 && pacman.xv<0)	// If going left but can't carry on that way, stop him.
		pacman.xv=0;
	if (r==0 && pacman.xv>0)	// If going right but can't carry on that way, stop him.
		pacman.xv=0;
	if (u==0 && pacman.yv<0)	// If going up but can't carry on that way, stop him.
		pacman.yv=0;
	if (d==0 && pacman.yv>0)	// If going down but can't carry on that way, stop him.
		pacman.yv=0;

	if (jlr==-1 && l)		// Want to go left and can ?
	{
		pacman.dir=PM_LEFT;
		pacman.xv=-PACMAN_VEL;
		pacman.yv=0;
	}
	else if (jlr==1 && r)	// Want to go right and can ?
	{
		pacman.dir=PM_RIGHT;
		pacman.xv=PACMAN_VEL;
		pacman.yv=0;
	}
	else if (jud==-1 && u)	// Want to go up and can ?
	{
		pacman.dir=PM_UP;
		pacman.yv=-PACMAN_VEL;
		pacman.xv=0;
	}
	else if (jud==1 && d)	// Want to go down and can ?
	{
		pacman.dir=PM_DOWN;
		pacman.yv=PACMAN_VEL;
		pacman.xv=0;
	}

	pacman.x+=pacman.xv;
	pacman.y+=pacman.yv;

	jlr=jud=0;	// Reset joypad press.

	if (Map[((POS(pacman.x)+8)>>3)+(((POS(pacman.y)+8)>>3)<<5)] >=1 &&
		Map[((POS(pacman.x)+8)>>3)+(((POS(pacman.y)+8)>>3)<<5)] <= 3)	// Eating a pill ?
	{
		if (Map[((POS(pacman.x)+8)>>3)+(((POS(pacman.y)+8)>>3)<<5)] == 3)	// Eating a POWER PILL ?
		{
			PillsEaten++;
			Score+=50;
			for (udx=0;udx<4;udx++)
			{
				if ((ghost[udx].Flags&EATEN)==0)	// Don't turn blue if already eaten.
				{
					ghost[udx].BlueTimer = (20-(GhostLevel>>1))*60;	// Between 20 and 8 seconds, depending on level.
					ghost[udx].Flags |= QUICK_TURN;
					ghost[udx].xv>>=1;
					ghost[udx].dxv>>=1;
					ghost[udx].yv>>=1;
					ghost[udx].dyv>>=1;
				}
			}
		}
		else
		{
			PillsEaten++;
			Score+=10;
		}
		Map[((POS(pacman.x)+8)>>3)+(((POS(pacman.y)+8)>>3)<<5)] = 0;	// Clear dot from detection map.
		m0[((POS(pacman.x)+8)>>3)+(((POS(pacman.y)+8)>>3)<<5)] = 0;		// Clear dot from screen.
	}

	if (POS(pacman.x)<=-16)		// Cater for tunnel left.
		pacman.x = TOFIXED(239);
	else if (POS(pacman.x)>=240)	// Cater for tunnel right.
		pacman.x = TOFIXED(-15);

	if (PillsEaten == NUM_PILLS)
		ExitLevel=NEXT_LEVEL;

}


void	AnimatePacMan(void)
{
	u32	*src;
	int	x;

	DMA3Call((u32)(&_chr)+(PillAnims[(vblcounter&15)>>2]*32), (u32)&tiles[3*32], 0x84000008);		// Animate pill;

	src = &sprite;

	if (pacman.Timer > -1)
	{
		pacman.Timer++;
		if (pacman.Timer < 60)	// Wait 1 second before doing dying anim.
			src += PMFrames[pacman.dir][(pacman.def>>1)&3]<<5;
		else
		{
			if (pacman.Timer == 60)	// Remove ghost sprites after 1 second.
			{
				for(x=0;x<4;x++)
					ghost[x].y = TOFIXED(160+ScrollPos);
				pacman.dir = PM_DOWN;
			}
			x = pacman.Timer-60;
			x>>=3;
			if (x>11)	// Finished death anim so restart player.
			{
				ExitLevel=DEATH;
				pacman.y = TOFIXED(160+ScrollPos);
				SetSpritePacman();		// Draw Pacman.
			}
			else
				src += (PACMAN5+x)<<5;
		}
	}
	else
	{
		if (pacman.xv||pacman.yv)	// moving, so animate.
			pacman.def++;
		src += PMFrames[pacman.dir][(pacman.def>>1)&3]<<5;
	}

	DMA3Call((u32)src, (u32)OBJCHARDATA, 0x84000020);		// Copy Sprites Character Bloc to VRAM
}

/*************************************************************
 *
 *
 *
 *
 *
 **************************************************************/

// read joypad... :)

void ReadJoypad(void) 									//joypad checking routine.
{
  	RawEdge = RawPad;
	RawPad = (~P1) & 0x3ff;				// Raw values. (NOT'ed so that 1 is on! Nintendo! What are they on?)
  	RawEdge = (RawEdge^RawPad)&RawPad;	// Debounce values.
}

void	download_sprite_data(void)
{
	SetPaletteSpr((u16 *)&_pal, 0, 64);
	DMA3Call((u32)&sprite, (u32)OBJCHARDATA, 0x84001330);		// Copy Sprites Character Bloc to VRAM
}

void DMA3Call(u32 Src,u32 Dst,u32 Cnt)
{
	DMA3SAD = Src;	// where the data is
	DMA3DAD = Dst;	// where the data goes
	DMA3CNT = Cnt;	// how much data should be loaded
}

void SetSpritePacman(void)
{
	int	nx,ny;

	nx=POS(pacman.x)+PMOffsets[pacman.dir][0];	// Offset for x 'cos of sprite flip.
	ny=POS(pacman.y)-ScrollPos+1+PMOffsets[pacman.dir][1];	// Offset for y 'cos of sprite flip.
	OAM[PACMAN].Attrib0 = (ny&255);	// Setup y.
	OAM[PACMAN].Attrib1 = 0x4000|(nx&511)|PMFrames[pacman.dir][4];	// Setup x and flips.
	OAM[PACMAN].Attrib2 = (3<<10);
}

void SetSpriteGhost(void)
{
	GHOSTS 	*g;
	int	x;

	g=&ghost[0];

	for (x=0;x<4;x++)
	{
		OAM[GHOST1+x].Attrib0 = (POS(g->y)-ScrollPos)&255;
		OAM[GHOST1+x].Attrib1 = 0x4000|((POS(g->x))&511);
		if (g->Flags&EATEN)
			OAM[GHOST1+x].Attrib2 = (SPR(GHOSTEYEUP)+(g->dir<<2))|(3<<10)|(x<<12);
		else if (g->BlueTimer > 4*60 || (g->BlueTimer && g->BlueTimer < (4*60) && (vblcounter&4)))
			OAM[GHOST1+x].Attrib2 = (SPR(GHOSTRUN1)+((vblcounter)&4))|(3<<10)|(x<<12);
		else
			OAM[GHOST1+x].Attrib2 = (SPR(GHOSTUP1)+((vblcounter)&4)+(g->dir<<3))|(3<<10)|(x<<12);
		g++;
	}
}

void SetupBonus()
{
	SCORES	*s;

	if (BonusTimer==0)
		return;

	if (--BonusTimer)
		return;

	s = &scorespr;

	s->timer=10*60+(GetRandom()&511);
	s->x = TOFIXED(112);
	s->y = TOFIXED(132);
	s->def = BONUS1+(CurrentLevel);
}

// Sprite routine for bonus sprites

void SetSpriteBonus(void)
{
	SCORES	*s;

	s = &scorespr;

	if (s->timer!=-1)
	{
		s->timer--;

		if (s->timer==-1)
		{
			s->y = TOFIXED(160+ScrollPos);	// Push it off screen.
			BonusTimer=(30*60)+(GetRandom()&1023);	// Between 30 and 60 seconds bonus repop.
		}

		if (ABS(POS(pacman.x)-POS(s->x))<8&&
			ABS(POS(pacman.y)-POS(s->y))<8)	// Eaten the bonus fruit?
		{
			Score+=BonusScores[CurrentLevel];
			s->timer=-1;
			OAM[GHOST1SCORE+1].Attrib0 = 160;	// Remove bonus from screen.
			DoGhostScore((GHOSTS *)0,BONUS1SC+CurrentLevel);
			BonusTimer=(30*60)+(GetRandom()&1023);	// Between 30 and 60 seconds bonus repop.
		}
		else
		{
			OAM[GHOST1SCORE+1].Attrib0 = POS(s->y)-ScrollPos;
			OAM[GHOST1SCORE+1].Attrib1 = 0x4000|POS(s->x);
		 	OAM[GHOST1SCORE+1].Attrib2 = SPR(s->def)|(2<<10);
		}
	}
}

void DoGhostScore(GHOSTS *g,int def)
{
	int	time=1*60,x,y;

	if (def<BONUS1SC)	// Eaten a ghost?
	{
		OAM[GHOST1+(g-ghost)].Attrib0 = 160;
		x = POS(g->x)&511;
		y = POS(g->y)&255;
	}
	else
	{
		x=112;
		y=132;
	}

	OAM[GHOST1SCORE].Attrib0 = y-ScrollPos;
	OAM[GHOST1SCORE].Attrib1 = 0x4000|x;
	OAM[GHOST1SCORE].Attrib2 = SPR(def)|(2<<10);
	while (time--)
	{
		VSync(1);
	}
	OAM[GHOST1SCORE].Attrib0 = 160;
}

void	_VBlankIntr(void)
{
	vblcounter++;
	vblcounter&=0x7fff;
}

// VCount irq!
void	_VCountIntr(void)
{
}

void SetPaletteChr(u16 *SrcPal, int StartCol, int NumCols)
{
	u16 *DestPal;

	DestPal = CHR_PALETTE+StartCol;
	while (NumCols--)
		*DestPal++ = *SrcPal++;
}

void SetPaletteSpr(u16 *SrcPal, int StartCol, int NumCols)
{
	SetPaletteChr(SrcPal,StartCol+256,NumCols);	// Sprite Palette is directly after BG palette! Add 256!!
}

int rand_1 = 0xd371f947;
int GetRandom(void)
{
	rand_1 = rand_1 * 0x41c64e6d + 0x3039;
	return( ( rand_1 >> 16 )&0x7fff );
}



void DisplayScore()
{
	char buf[32],i,l,x;

	for (i=x=0;i<8;i++,x+=32)	// Clear bottom line.
	{
		OAM[TxtSpr].Attrib0 = 152|0x4000;
		OAM[TxtSpr].Attrib1 = x|0x4000;
		OAM[TxtSpr].Attrib2 = COVER_SPR;
		TxtSpr--;
	}

	if (P1Timer)
	{
		P1Timer--;
		sprintf(buf,"PLAYER ONE");
		PrintString(120,40,buf,TXT_CENTRE);
	}
	else if (ReadyTimer)
	{
		ReadyTimer--;
		sprintf(buf,"READY!");
		PrintString(120,40,buf,TXT_CENTRE);
	}

	sprintf(buf,"%d",HighScore);
	PrintString(120,152,buf,TXT_CENTRE);
	sprintf(buf,"%d",Score);
	PrintString(8,152,buf,0);

	if (pacman.Lives)
	{
		l=(pacman.Lives<5) ? pacman.Lives:5;
		for (i=0;i<l;i++)
			buf[i]='@';

		if (pacman.Lives>5)
			buf[i++]='!';

		buf[i]=0;
		PrintString(232-(i<<3),152,buf,0);
	}
}

void ClearRemainingTextSprites()
{
	while(TxtSpr>64)
	{
		OAM[TxtSpr].Attrib0 = 160;
		TxtSpr--;
	}
}

void VSync(int nframes)
{
	int	sync;

	while (nframes)
	{
 		sync=vblcounter;
 		while (sync==vblcounter)
			;
		nframes--;
	}
}

void Fade()
{
	if (FadeSrc!=FadeDst)
	{
		if (FadeSrc > FadeDst)
		{
			FadeSrc -= FadeSpeed;
			if (FadeSrc < FadeDst)
				FadeSrc=FadeDst;
		}
		else
		{
			FadeSrc += FadeSpeed;
			if (FadeSrc > FadeDst)
				FadeSrc=FadeDst;
		}

		if (FadeSrc > 128)	// White Fader?
		{
		 	BLDMOD = FadeScreens|(2<<6);
			COLY=(FadeSrc-128)>>3;
		}
		else
		{
			BLDMOD = FadeScreens|(3<<6);
			COLY=(128-FadeSrc)>>3;
		}
	}
}

void WaitFade()
{
	while (FadeSrc!=FadeDst)
	{
		VSync(1);
		Fade();
	}
}

void PauseGame()
{
	FadeDst=96;	// Fade down half way.

	while (1)
	{
		VSync(1);
		Fade();
		ReadJoypad();

		if (vblcounter & 16)
		{
			OAM[TxtSpr].Attrib0 = 0x4000|60;	// Setup y.
			OAM[TxtSpr].Attrib1 = 0xc000|96;	// Setup x and flips.
			OAM[TxtSpr].Attrib2 = PAUSE_SPR;
		}
		else
			OAM[TxtSpr].Attrib0 = 160;

		if (RawEdge&J_START)
			break;
	}


	FadeDst=128;	// Fade down half way.
}

void GameOver()
{
	int	tmr=5*64;
	int	trns=0;
	int	scale=512+128;
	int	sin,cos,angle;

	angle=0;

	while (tmr--)
	{
		VSync(1);
		ReadJoypad();

		BLDMOD=16|((1+32)<<8);	// Set OBJ screen for colour effects (semi trans) against BG0.

		if (tmr < 64)
		{
			trns=tmr;
			scale+=12;
			angle-=16;
			angle&=1023;
		}
		else if (trns < 64)
		{
			angle+=16;
			angle&=1023;
			trns++;
			scale-=8;
		}

		OAM[TxtSpr].Attrib0 = 0x4000|(60-16)|(1<<10)|(3<<8);	// Setup y, shape, trans, scale and double size.
		OAM[TxtSpr].Attrib1 = 0xc000|(88-32);	// Setup x and flips.
		OAM[TxtSpr].Attrib2 = GAMEOVER_SPR;
		COLEV = (trns>>2)|(((64-trns)>>2)<<8);

		sin = (SIN(angle)*scale)>>7;
		cos = (COS(angle)*scale)>>7;

		OAM[0].RotateScale = cos;
		OAM[1].RotateScale = sin;
		OAM[2].RotateScale = -sin;
		OAM[3].RotateScale = cos;

		if (RawEdge&J_SELECT)
			break;
	}

	OAM[TxtSpr].Attrib0 = 160;

	FadeDst=0;	// Fade down to black.
	WaitFade();
}

void FinishLevel()
{
	int	x;

	VSync(30);	//Wait half a second before removing sprites.

	for (x=0;x<5;x++)
		OAM[x].Attrib0 = 160;

	vblcounter=0;

	x=8;
	while (x--)
	{
		VSync(32);
		SetPaletteChr(&_pal+((vblcounter&32)<<1),0,16);	// Upload palette into palette ram.
	}

	VSync(1);
	SetPaletteChr(&_pal,0,16);	// Upload palette into palette ram.
}

void ProfileRGB(short r,short g, short b)
{
	u16 *DestPal;
	DestPal = CHR_PALETTE;
	*DestPal = (r>>3)|((g>>3)<<5)|((b>>3)<<10);
}

